statemachine("Workflow for state", "State").addRequirements(requirements([{name: "Issue", fields: [{name: "Subtask", type: {name: "IssueLinkPrototype", multiple: true, fields: [{name: "parent for", type: {name: "Issue", multiple: true}}, {name: "subtask of", type: {name: "Issue", multiple: true}}]}}, {name: "Type", type: {name: "EnumField", values: ["Hotfix"]}}, {name: "summary", type: {name: "string", primitive: true}}, {name: "project", type: {name: "Project", fields: [{name: "leader", type: {name: "User", methods: [{name: "notify", returnType: {name: "void"}, paramTypes: [{name: "string"}, {name: "string"}]}]}}]}}, {name: "Assignee", type: {name: "User", methods: [{name: "notify", returnType: {name: "void"}, paramTypes: [{name: "string"}, {name: "string"}]}]}}, {name: "Branch", type: {name: "EnumField", values: ["Ticket", "Parent", "Master"]}}, {name: "Tester", type: {name: "User"}}, {name: "Code reviewer", type: {name: "User"}}, {name: "State", type: {name: "State", values: ["Submitted", "Accepted", "Can't Reproduce", "Won't fix", "Duplicate", "Open", "In Progress", "Code revision", "Fixed", "Verified"]}}]}]));
from("Submitted").on("send to backlog").transitTo("Accepted");
from("Submitted").on("start hotfix").perform(function(ctx) {
  if (safeCall(safeCall(ctx.issue,"get", ["subtask of"]),"isNotEmpty", [], false)) {
    // feature subtasks are created directly in open state on feature dashboard and this workflow is executed
    // we just don't want to mark them as hotfixes
  } else {
    // otherwise from submitted we can only open hotfixes
    safeCall(ctx.issue,"set", ["Type", find("Hotfix")], null);
    if (equals(safeCall(ctx.issue,"get", ["Assignee"]), null)) {
      invoke(ctx, safeCall(safeCall(ctx.issue,"get", ["project"]),"get", ["leader"]), "notify", ["[ASSIGNEE] " + invoke(ctx, ctx.issue, "getId", []) + ": " + safeCall(ctx.issue,"get", ["summary"], null),invoke(ctx, ctx.issue, "getUrl", [])]);
    }
  }
}).transitTo("Open");
from("Submitted").on("can't reproduce").transitTo("Can't Reproduce");
from("Submitted").on("won't fix").transitTo("Won't fix");
from("Submitted").on("duplicate").transitTo("Duplicate");
from("Accepted").on("start hotfix").perform(function(ctx) {
  safeCall(ctx.issue,"set", ["Type", find("Hotfix")], null);
  if (equals(safeCall(ctx.issue,"get", ["Assignee"]), null)) {
    invoke(ctx, safeCall(safeCall(ctx.issue,"get", ["project"]),"get", ["leader"]), "notify", ["[ASSIGNEE] " + invoke(ctx, ctx.issue, "getId", []) + ": " + safeCall(ctx.issue,"get", ["summary"], null),invoke(ctx, ctx.issue, "getUrl", [])]);
  }
}).transitTo("Open");
from("Accepted").on("send to development").perform(function(ctx) {
  if (equals(safeCall(ctx.issue,"get", ["Assignee"]), null)) {
    invoke(ctx, safeCall(safeCall(ctx.issue,"get", ["project"]),"get", ["leader"]), "notify", ["[ASSIGNEE] " + invoke(ctx, ctx.issue, "getId", []) + ": " + safeCall(ctx.issue,"get", ["summary"], null),invoke(ctx, ctx.issue, "getUrl", [])]);
  }
}).transitTo("Open");
from("Accepted").on("can't reproduce").transitTo("Can't Reproduce");
from("Accepted").on("won't fix").transitTo("Won't fix");
from("Accepted").on("duplicate").transitTo("Duplicate");
from("Can't Reproduce").on(" reopen ").transitTo("Submitted");
from("Won't fix").on(" reopen ").transitTo("Submitted");
from("Duplicate").on(" reopen ").transitTo("Submitted");
from("Open").on("start work").perform(function(ctx) {
  assert(ctx, equals(ctx.loggedInUser, safeCall(ctx.issue,"get", ["Assignee"])),"Only assignee can start work on this ticket.");
  if (equals(safeCall(ctx.issue,"get", ["Type"]), find("Hotfix"))) {
    message(ctx,"Don't forget to create ticket branch from master for this ticket.");
  } else if (equals(safeCall(ctx.issue,"get", ["Branch"]), find("Ticket"))) {
    message(ctx,"Don't forget to create ticket branch from integration for this ticket.");
  }
}).transitTo("In Progress");
from("Open").on("can't reproduce").transitTo("Can't Reproduce");
from("Open").on("won't fix").transitTo("Won't fix");
from("Open").on("duplicate").transitTo("Duplicate");
from("In Progress").on("send to code revision").perform(function(ctx) {
  if (!equals(safeCall(ctx.issue,"get", ["Branch"]), find("Parent"))) {
    assert(ctx, equals(ctx.loggedInUser, safeCall(ctx.issue,"get", ["Assignee"])),"Only assignee can finish work on this ticket.");
    assert(ctx, equals(safeCall(ctx.issue,"get", ["Tester"]), null),"This ticket is already in testing stage and shouldn't be sent to code revision anymore. You should send it for test.");
    assert(ctx, equals(safeCall(ctx.issue,"get", ["Branch"]), find("Ticket")),"Only tickets in ticket branch can be send to code revision.");
    if (equals(safeCall(ctx.issue,"get", ["Code reviewer"]), null)) {
      invoke(ctx, safeCall(safeCall(ctx.issue,"get", ["project"]),"get", ["leader"]), "notify", ["[CODE REVIEWER] " + invoke(ctx, ctx.issue, "getId", []) + ": " + safeCall(ctx.issue,"get", ["summary"], null),invoke(ctx, ctx.issue, "getUrl", [])]);
    }
  } else {
    // workflow within the feature is free
    if (equals(safeCall(ctx.issue,"get", ["Code reviewer"]), null)) {
      invoke(ctx, safeCall(safeCall(safeCall(ctx.issue,"get", ["subtask of"]),"first", []),"get", ["Assignee"]), "notify", ["[CODE REVIEWER] " + invoke(ctx, ctx.issue, "getId", []) + ": " + safeCall(ctx.issue,"get", ["summary"], null),invoke(ctx, ctx.issue, "getUrl", [])]);
    }
  }
}).transitTo("Code revision");
from("In Progress").on("fix it ").perform(function(ctx) {
  if (!equals(safeCall(ctx.issue,"get", ["Branch"]), find("Parent"))) {
    assert(ctx, equals(ctx.loggedInUser, safeCall(ctx.issue,"get", ["Assignee"])),"Only assignee can finish work on this ticket.");
    if (!equals(safeCall(ctx.issue,"get", ["Type"]), find("Hotfix"))) {
      assert(ctx, !equals(safeCall(ctx.issue,"get", ["Branch"]), find("Ticket")),"Tickets in ticket branch are not prepared for testing yet. Please merge this ticket to integration branch " + "and change Branch field of this ticket.");
      if (equals(safeCall(ctx.issue,"get", ["Tester"]), null)) {
        invoke(ctx, safeCall(safeCall(ctx.issue,"get", ["project"]),"get", ["leader"]), "notify", ["[TESTER] " + invoke(ctx, ctx.issue, "getId", []) + ": " + safeCall(ctx.issue,"get", ["summary"], null),invoke(ctx, ctx.issue, "getUrl", [])]);
      }
    } else {
      safeCall(ctx.issue,"set", ["Branch", find("Master")], null);
      message(ctx,"Done hotfixes are moved instantly to master to be released. Please merge ticket branch to master immediately!");
    }
    // workflow within the feature is free
  }
}).transitTo("Fixed");
from("In Progress").on("can't reproduce").transitTo("Can't Reproduce");
from("In Progress").on("won't fix").transitTo("Won't fix");
from("In Progress").on("duplicate").transitTo("Duplicate");
from("Code revision").on(" reopen ").perform(function(ctx) {
  if (!equals(safeCall(ctx.issue,"get", ["Branch"]), find("Parent"))) {
    assert(ctx, equals(ctx.loggedInUser, safeCall(ctx.issue,"get", ["Code reviewer"])),"Only code reviewer can reopen this ticket.");
    assert(ctx, equals(safeCall(ctx.issue,"get", ["Branch"]), find("Ticket")),"Only tickets in ticket branch can be sent back from code revision.");
  }
  // workflow within the feature is free
}).transitTo("In Progress");
from("Code revision").on("send to integration").perform(function(ctx) {
  if (!equals(safeCall(ctx.issue,"get", ["Branch"]), find("Parent"))) {
    assert(ctx, equals(ctx.loggedInUser, safeCall(ctx.issue,"get", ["Code reviewer"])),"Only code reviewer can finish this ticket.");
    assert(ctx, equals(safeCall(ctx.issue,"get", ["Branch"]), find("Ticket")),"Only tickets in ticket branch can be sent to integration.");
    if ((!equals(safeCall(ctx.issue,"get", ["Type"]), find("Hotfix")))) {
      invoke(ctx, ctx.issue, "addComment", ["Code revision passed, please merge this ticket to integration branch and change branch to Integration."]);
    } else {
      invoke(ctx, ctx.issue, "addComment", ["Code revision passed, please merge this ticket to master branch and change branch to Master."]);
    }
  }
  // workflow within the feature is free
}).transitTo("Fixed");
from("Fixed").on("reopen").perform(function(ctx) {
  if (!equals(safeCall(ctx.issue,"get", ["Branch"]), find("Parent"))) {
    assert(ctx, equals(ctx.loggedInUser, safeCall(ctx.issue,"get", ["Tester"])),"Only tester can reopen this ticket.");
    assert(ctx, !equals(safeCall(ctx.issue,"get", ["Branch"]), find("Ticket")),"Fixed tickets in ticket branch should be merged to integration, not reopened.");
  }
  // workflow within the feature is free
}).transitTo("In Progress");
from("Fixed").on("tests passed").perform(function(ctx) {
  if (!equals(safeCall(ctx.issue,"get", ["Branch"]), find("Parent"))) {
    if ((!equals(safeCall(ctx.issue,"get", ["Type"]), find("Hotfix")))) {
      assert(ctx, equals(ctx.loggedInUser, safeCall(ctx.issue,"get", ["Tester"])),"Only tester can mark this ticket verified.");
    }
    assert(ctx, !equals(safeCall(ctx.issue,"get", ["Branch"]), find("Ticket")),"Tickets in ticket branch are not sent for tests yet, so they cannot be marked as verified.");
  }
  // workflow within the feature is free
}).transitTo("Verified");
from("Verified");